import numpy as np
import pytest
from astropy import cosmology as cosmo

from autolens import dimensions as dim
from autolens.data.array import grids
from autolens.data.array import mask as msk
from autolens.model.profiles import mass_profiles as mp

from test.unit.mock.model import mock_cosmology

grid = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [2.0, 4.0]])


class TestAbstractNFW(object):
    def test__coord_function_f__correct_values(self):
        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=3.0
        )

        # r > 1

        assert truncated_nfw.coord_func_f(
            grid_radius=np.array([2.0, 3.0])
        ) == pytest.approx(np.array([0.604599, 0.435209]), 1.0e-4)

        # r < 1
        assert truncated_nfw.coord_func_f(
            grid_radius=np.array([0.5, 1.0 / 3.0])
        ) == pytest.approx(1.52069, 1.86967, 1.0e-4)
        #
        # r == 1
        assert (
            truncated_nfw.coord_func_f(grid_radius=np.array([1.0, 1.0]))
            == np.array([1.0, 1.0])
        ).all()

    def test__coord_function_g__correct_values(self):
        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=3.0
        )

        # r > 1

        assert truncated_nfw.coord_func_g(
            grid_radius=np.array([2.0, 3.0])
        ) == pytest.approx(np.array([0.13180, 0.070598]), 1.0e-4)

        # r < 1

        assert truncated_nfw.coord_func_g(
            grid_radius=np.array([0.5, 1.0 / 3.0])
        ) == pytest.approx(np.array([0.69425, 0.97838]), 1.0e-4)

        # r == 1
        assert (
            truncated_nfw.coord_func_g(grid_radius=np.array([1.0, 1.0]))
            == np.array([1.0 / 3.0, 1.0 / 3.0])
        ).all()

    def test__coord_function_h__correct_values(self):
        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=3.0
        )

        assert truncated_nfw.coord_func_h(
            grid_radius=np.array([0.5, 3.0])
        ) == pytest.approx(np.array([0.134395, 0.840674]), 1.0e-4)

    def test__coord_function_k__correct_values(self):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=2.0
        )

        assert truncated_nfw.coord_func_k(
            grid_radius=np.array([2.0, 3.0])
        ) == pytest.approx(np.array([-0.09983408, -0.06661738]), 1.0e-4)

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=4.0
        )

        assert truncated_nfw.coord_func_k(
            grid_radius=np.array([2.0, 3.0])
        ) == pytest.approx(np.array([-0.19869011, -0.1329414]), 1.0e-4)

    def test__coord_function_l__correct_values(self):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=2.0
        )

        assert truncated_nfw.coord_func_l(
            grid_radius=np.array([2.0, 2.0])
        ) == pytest.approx(np.array([0.00080191, 0.00080191]), 1.0e-4)

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=3.0
        )

        assert truncated_nfw.coord_func_l(
            grid_radius=np.array([2.0, 2.0])
        ) == pytest.approx(np.array([0.00178711, 0.00178711]), 1.0e-4)

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=3.0
        )
        assert truncated_nfw.coord_func_l(
            grid_radius=np.array([3.0, 3.0])
        ) == pytest.approx(np.array([0.00044044, 0.00044044]), 1.0e-4)

    def test__coord_function_m__correct_values(self):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=2.0
        )

        assert truncated_nfw.coord_func_m(
            grid_radius=np.array([2.0, 2.0])
        ) == pytest.approx(np.array([0.0398826, 0.0398826]), 1.0e-4)

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=3.0
        )

        assert truncated_nfw.coord_func_m(
            grid_radius=np.array([2.0, 2.0])
        ) == pytest.approx(np.array([0.06726646, 0.06726646]), 1.0e-4)

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=3.0
        )

        assert truncated_nfw.coord_func_m(
            grid_radius=np.array([3.0, 3.0])
        ) == pytest.approx(np.array([0.06946888, 0.06946888]), 1.0e-4)

    def test__rho_at_scale_radius__numerical_values_in_default_units(self):
        cosmology = mock_cosmology.MockCosmology(
            kpc_per_arcsec=2.0, critical_surface_density=2.0
        )

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)
        rho = nfw.rho_at_scale_radius_for_units(
            redshift_profile=0.5, redshift_source=1.0, cosmology=cosmology
        )
        assert rho == pytest.approx(1.0, 1e-3)

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=3.0, scale_radius=1.0)
        rho = nfw.rho_at_scale_radius_for_units(
            redshift_profile=0.5, redshift_source=1.0, cosmology=cosmology
        )
        assert rho == pytest.approx(3.0, 1e-3)

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=4.0)
        rho = nfw.rho_at_scale_radius_for_units(
            redshift_profile=0.5, redshift_source=1.0, cosmology=cosmology
        )
        assert rho == pytest.approx(0.25, 1e-3)

    def test__rho_at_scale_radius__unit_conversions(self):
        cosmology = mock_cosmology.MockCosmology(
            arcsec_per_kpc=0.5, kpc_per_arcsec=2.0, critical_surface_density=2.0
        )

        nfw_arcsec = mp.SphericalNFW(
            centre=(dim.Length(0.0, "arcsec"), dim.Length(0.0, "arcsec")),
            kappa_s=1.0,
            scale_radius=dim.Length(1.0, "arcsec"),
        )

        rho = nfw_arcsec.rho_at_scale_radius_for_units(
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert rho == pytest.approx(2.0, 1e-3)

        # When converting to kpc, the critical surface density is divided by kpc_per_arcsec**2.0 = 2.0**2.0
        # The scale radius also becomes scale_radius*kpc_per_arcsec = 2.0

        rho = nfw_arcsec.rho_at_scale_radius_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert rho == pytest.approx(0.5 / 2.0, 1e-3)

        rho = nfw_arcsec.rho_at_scale_radius_for_units(
            unit_length="arcsec",
            unit_mass="angular",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert rho == pytest.approx(1.0, 1e-3)

        # This will make the critical surface density 4.0, with the same conversions as above

        cosmology = mock_cosmology.MockCosmology(
            arcsec_per_kpc=0.25, kpc_per_arcsec=4.0, critical_surface_density=2.0
        )

        rho = nfw_arcsec.rho_at_scale_radius_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert rho == pytest.approx(0.5 / 4.0, 1e-3)

        cosmology = mock_cosmology.MockCosmology(
            arcsec_per_kpc=0.25, kpc_per_arcsec=4.0, critical_surface_density=4.0
        )

        rho = nfw_arcsec.rho_at_scale_radius_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert rho == pytest.approx(0.25 / 4.0, 1e-3)

        nfw_kpc = mp.SphericalNFW(
            centre=(dim.Length(0.0, "kpc"), dim.Length(0.0, "kpc")),
            kappa_s=1.0,
            scale_radius=dim.Length(1.0, "kpc"),
        )

        cosmology = mock_cosmology.MockCosmology(
            arcsec_per_kpc=0.5, kpc_per_arcsec=2.0, critical_surface_density=2.0
        )
        rho = nfw_kpc.rho_at_scale_radius_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert rho == pytest.approx(0.5, 1e-3)

        rho = nfw_kpc.rho_at_scale_radius_for_units(
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert rho == pytest.approx(2.0 / 0.5, 1e-3)

        rho = nfw_kpc.rho_at_scale_radius_for_units(
            unit_length="kpc",
            unit_mass="angular",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )
        assert rho == pytest.approx(1.0, 1e-3)

    def test__delta_concentration_value_in_default_units(self):
        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)

        cosmology = mock_cosmology.MockCosmology(
            arcsec_per_kpc=1.0,
            kpc_per_arcsec=1.0,
            critical_surface_density=1.0,
            cosmic_average_density=1.0,
        )

        delta_concentration = nfw.delta_concentration_for_units(
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="solMass",
            cosmology=cosmology,
        )
        assert delta_concentration == pytest.approx(1.0, 1e-3)

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=3.0, scale_radius=1.0)
        delta_concentration = nfw.delta_concentration_for_units(
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="solMass",
            cosmology=cosmology,
        )
        assert delta_concentration == pytest.approx(3.0, 1e-3)

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=4.0)
        delta_concentration = nfw.delta_concentration_for_units(
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_mass="solMass",
            cosmology=cosmology,
        )
        assert delta_concentration == pytest.approx(0.25, 1e-3)

        # cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.5, kpc_per_arcsec=2.0, critical_surface_density=2.0,
        #                           cosmic_average_density=1.0)
        #
        # nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)
        # delta_concentration = nfw.delta_concentration(unit_length='kpc', unit_mass='solMass', redshift_lens=0.5,
        #                                               redshift_source=1.0, cosmology=cosmology)
        # assert delta_concentration == pytest.approx(0.5, 1e-3)
        #
        # nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=2.0, scale_radius=1.0)
        # delta_concentration = nfw.delta_concentration(unit_length='kpc', unit_mass='solMass', redshift_lens=0.5, redshift_source=1.0,
        #                                               cosmology=cosmology)
        # assert delta_concentration == pytest.approx(1.0, 1e-3)
        #
        # nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=2.0, scale_radius=20.0)
        # delta_concentration = nfw.delta_concentration(unit_length='kpc',  unit_mass='solMass', redshift_lens=0.5, redshift_source=1.0,
        #                                               cosmology=cosmology)
        # assert delta_concentration == pytest.approx(0.05, 1e-3)
        #
        # cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.5, kpc_per_arcsec=2.0, critical_surface_density=2.0,
        #                           cosmic_average_density=2.0)
        #
        # nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)
        # delta_concentration = nfw.delta_concentration(unit_length='kpc',  unit_mass='solMass', redshift_lens=0.5, redshift_source=1.0,
        #                                               cosmology=cosmology)
        # assert delta_concentration == pytest.approx(0.25, 1e-3)

    def test__solve_concentration(self):
        cosmology = mock_cosmology.MockCosmology(
            arcsec_per_kpc=1.0,
            kpc_per_arcsec=1.0,
            critical_surface_density=1.0,
            cosmic_average_density=1.0,
        )

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)

        concentration = nfw.concentration_for_units(
            redshift_profile=0.5, redshift_source=1.0, cosmology=cosmology
        )

        assert concentration == pytest.approx(0.0074263, 1.0e-4)

    def test__radius_at_200__different_length_units_include_conversions(self):
        nfw_arcsec = mp.SphericalNFW(
            centre=(dim.Length(0.0, "arcsec"), dim.Length(0.0, "arcsec")),
            kappa_s=1.0,
            scale_radius=dim.Length(1.0, "arcsec"),
        )

        cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.2, kpc_per_arcsec=5.0)

        concentration = nfw_arcsec.concentration_for_units(
            cosmology=cosmology, redshift_profile=0.5, redshift_source=1.0
        )

        radius_200 = nfw_arcsec.radius_at_200_for_units(
            unit_length="arcsec",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        assert radius_200 == concentration * 1.0

        cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.5, kpc_per_arcsec=2.0)

        concentration = nfw_arcsec.concentration_for_units(
            unit_length="kpc",
            cosmology=cosmology,
            redshift_profile=0.5,
            redshift_source=1.0,
        )

        radius_200 = nfw_arcsec.radius_at_200_for_units(
            unit_length="kpc",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        assert radius_200 == 2.0 * concentration * 1.0

        nfw_kpc = mp.SphericalNFW(
            centre=(dim.Length(0.0, "kpc"), dim.Length(0.0, "kpc")),
            kappa_s=1.0,
            scale_radius=dim.Length(1.0, "kpc"),
        )

        cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.2, kpc_per_arcsec=5.0)

        concentration = nfw_kpc.concentration_for_units(
            unit_length="kpc",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        radius_200 = nfw_kpc.radius_at_200_for_units(
            unit_length="kpc",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        assert radius_200 == concentration * 1.0

        cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.5, kpc_per_arcsec=2.0)

        concentration = nfw_kpc.concentration_for_units(
            unit_length="arcsec",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        radius_200 = nfw_kpc.radius_at_200_for_units(
            unit_length="arcsec",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        assert radius_200 == concentration * 1.0 / 2.0

    def test__mass_at_200__unit_conversions_work(self):
        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)

        cosmology = mock_cosmology.MockCosmology(
            arcsec_per_kpc=1.0,
            kpc_per_arcsec=1.0,
            critical_surface_density=1.0,
            cosmic_average_density=1.0,
        )

        radius_at_200 = nfw.radius_at_200_for_units(
            unit_length="arcsec",
            redshift_profile=0.5,
            redshift_source=1.0,
            cosmology=cosmology,
        )

        mass_at_200 = nfw.mass_at_200_for_units(
            cosmology=cosmology,
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_length="arcsec",
            unit_mass="solMass",
        )

        mass_calc = (
            200.0
            * ((4.0 / 3.0) * np.pi)
            * cosmology.cosmic_average_density
            * (radius_at_200 ** 3.0)
        )
        assert mass_at_200 == pytest.approx(mass_calc, 1.0e-5)

        # cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.5, kpc_per_arcsec=2.0, critical_surface_density=2.0,
        #                           cosmic_average_density=1.0)
        #
        # radius_at_200 = nfw.radius_at_200_for_units(unit_length='arcsec', redshift_lens=0.5, redshift_source=1.0,
        #                                             cosmology=cosmology)
        #
        # mass_at_200 = nfw.mass_at_200(cosmology=cosmology, redshift_lens=0.5, redshift_source=1.0, unit_length='arcsec',
        #                               unit_mass='solMass')
        #
        # mass_calc = 200.0 * ((4.0 / 3.0) * np.pi) * cosmology.cosmic_average_density * (radius_at_200 ** 3.0)
        # assert mass_at_200 == pytest.approx(mass_calc, 1.0e-5)

    def test__values_of_quantities_for_real_cosmology__include_unit_conversions(self):

        cosmology = cosmo.LambdaCDM(H0=70.0, Om0=0.3, Ode0=0.7)

        nfw = mp.SphericalTruncatedNFW(
            kappa_s=0.5, scale_radius=5.0, truncation_radius=10.0
        )

        rho = nfw.rho_at_scale_radius_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            cosmology=cosmology,
        )

        delta_concentration = nfw.delta_concentration_for_units(
            redshift_profile=0.6,
            redshift_source=2.5,
            unit_length="kpc",
            unit_mass="solMass",
            redshift_of_cosmic_average_density="local",
            cosmology=cosmology,
        )

        concentration = nfw.concentration_for_units(
            redshift_profile=0.6,
            redshift_source=2.5,
            unit_length="kpc",
            unit_mass="solMass",
            redshift_of_cosmic_average_density="local",
            cosmology=cosmology,
        )

        radius_at_200 = nfw.radius_at_200_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="local",
            cosmology=cosmology,
        )

        mass_at_200 = nfw.mass_at_200_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="local",
            cosmology=cosmology,
        )

        mass_at_truncation_radius = nfw.mass_at_truncation_radius(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="local",
            cosmology=cosmology,
        )

        assert rho == pytest.approx(29027857.01622403, 1.0e-4)
        assert delta_concentration == pytest.approx(213451.19421263796, 1.0e-4)
        assert concentration == pytest.approx(18.6605624462417, 1.0e-4)
        assert radius_at_200 == pytest.approx(623.7751567997697, 1.0e-4)
        assert mass_at_200 == pytest.approx(27651532986258.375, 1.0e-4)
        assert mass_at_truncation_radius == pytest.approx(14877085957074.299, 1.0e-4)

        rho = nfw.rho_at_scale_radius_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            cosmology=cosmology,
        )

        delta_concentration = nfw.delta_concentration_for_units(
            redshift_profile=0.6,
            redshift_source=2.5,
            unit_length="kpc",
            unit_mass="solMass",
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        concentration = nfw.concentration_for_units(
            redshift_profile=0.6,
            redshift_source=2.5,
            unit_length="kpc",
            unit_mass="solMass",
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        radius_at_200 = nfw.radius_at_200_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        mass_at_200 = nfw.mass_at_200_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        mass_at_truncation_radius = nfw.mass_at_truncation_radius(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        assert rho == pytest.approx(29027857.01622403, 1.0e-4)
        assert delta_concentration == pytest.approx(110665.28111397651, 1.0e-4)
        assert concentration == pytest.approx(14.401574489517804, 1.0e-4)
        assert radius_at_200 == pytest.approx(481.40801817963467, 1.0e-4)
        assert mass_at_200 == pytest.approx(24516707575366.09, 1.0e-4)
        assert mass_at_truncation_radius == pytest.approx(13190486262169.797, 1.0e-4)

        nfw = mp.SphericalTruncatedNFW(
            kappa_s=0.5,
            scale_radius=dim.Length(3.0, "kpc"),
            truncation_radius=dim.Length(7.0, "kpc"),
        )

        rho = nfw.rho_at_scale_radius_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            cosmology=cosmology,
        )

        delta_concentration = nfw.delta_concentration_for_units(
            redshift_profile=0.6,
            redshift_source=2.5,
            unit_length="kpc",
            unit_mass="solMass",
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        concentration = nfw.concentration_for_units(
            redshift_profile=0.6,
            redshift_source=2.5,
            unit_length="kpc",
            unit_mass="solMass",
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        radius_at_200 = nfw.radius_at_200_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        mass_at_200 = nfw.mass_at_200_for_units(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        mass_at_truncation_radius = nfw.mass_at_truncation_radius(
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        assert rho == pytest.approx(323442484.90222085, 1.0e-4)
        assert delta_concentration == pytest.approx(1233086.3244882922, 1.0e-4)
        assert concentration == pytest.approx(36.61521013005619, 1.0e-4)
        assert radius_at_200 == pytest.approx(109.84563039016857, 1.0e-4)
        assert mass_at_200 == pytest.approx(291253092446.923, 1.0e-4)
        assert mass_at_truncation_radius == pytest.approx(177609204745.61484, 1.0e-4)

        rho = nfw.rho_at_scale_radius_for_units(
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            cosmology=cosmology,
        )

        delta_concentration = nfw.delta_concentration_for_units(
            redshift_profile=0.6,
            redshift_source=2.5,
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        concentration = nfw.concentration_for_units(
            redshift_profile=0.6,
            redshift_source=2.5,
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        radius_at_200 = nfw.radius_at_200_for_units(
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        mass_at_200 = nfw.mass_at_200_for_units(
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        mass_at_truncation_radius = nfw.mass_at_truncation_radius(
            unit_length="arcsec",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            cosmology=cosmology,
        )

        kpc_per_arcsec = 1.0 / cosmology.arcsec_per_kpc_proper(z=0.6).value

        assert rho == pytest.approx(323442484.90222085 * kpc_per_arcsec ** 3.0, 1.0e-4)
        assert delta_concentration == pytest.approx(1233086.3244882922, 1.0e-4)
        assert concentration == pytest.approx(36.61521013005619, 1.0e-4)
        assert radius_at_200 == pytest.approx(
            109.84563039016857 / kpc_per_arcsec, 1.0e-4
        )
        assert mass_at_200 == pytest.approx(291253092446.923, 1.0e-4)
        assert mass_at_truncation_radius == pytest.approx(177609204745.61484, 1.0e-4)


class TestGeneralizedNFW(object):
    def test__constructor_and_units(self):
        # gnfw = mp.EllipticalGeneralizedNFW(centre=(0.7, 1.0), axis_ratio=0.7, phi=45.0,
        #                                    kappa_s=2.0, inner_slope=1.5, scale_radius=10.0)
        #
        # assert gnfw.centre == (0.7, 1.0)
        # assert gnfw.axis_ratio == 0.7
        # assert gnfw.phi == 45.0
        # assert gnfw.kappa_s == 2.0
        # assert gnfw.inner_slope == 1.5
        # assert gnfw.scale_radius == 10.0

        gnfw = mp.SphericalGeneralizedNFW(
            centre=(1.0, 2.0), kappa_s=2.0, inner_slope=1.5, scale_radius=10.0
        )

        assert gnfw.centre == (1.0, 2.0)
        assert isinstance(gnfw.centre[0], dim.Length)
        assert isinstance(gnfw.centre[1], dim.Length)
        assert gnfw.centre[0].unit == "arcsec"
        assert gnfw.centre[1].unit == "arcsec"

        assert gnfw.axis_ratio == 1.0
        assert isinstance(gnfw.axis_ratio, float)

        assert gnfw.phi == 0.0
        assert isinstance(gnfw.phi, float)

        assert gnfw.kappa_s == 2.0
        assert isinstance(gnfw.kappa_s, float)

        assert gnfw.inner_slope == 1.5
        assert isinstance(gnfw.inner_slope, float)

        assert gnfw.scale_radius == 10.0
        assert isinstance(gnfw.scale_radius, dim.Length)
        assert gnfw.scale_radius.unit_length == "arcsec"

    # def test__coord_func_x_above_1(self):
    #     assert mp.EllipticalNFW.coord_func(2.0) == pytest.approx(0.60459, 1e-3)
    #
    #     assert mp.EllipticalNFW.coord_func(0.5) == pytest.approx(1.5206919, 1e-3)
    #
    #     assert mp.EllipticalNFW.coord_func(1.0) == 1.0

    def test__convergence_correct_values(self):
        gnfw = mp.SphericalGeneralizedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, inner_slope=1.5, scale_radius=1.0
        )
        assert gnfw.convergence_from_grid(grid=np.array([[2.0, 0.0]])) == pytest.approx(
            0.30840, 1e-3
        )

        gnfw = mp.SphericalGeneralizedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, inner_slope=1.5, scale_radius=1.0
        )
        assert gnfw.convergence_from_grid(grid=np.array([[2.0, 0.0]])) == pytest.approx(
            0.30840 * 2, 1e-3
        )

        # gnfw = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), kappa_s=1.0, axis_ratio=0.5,
        #                                    phi=90.0, inner_slope=1.5, scale_radius=1.0)
        # assert gnfw.convergence_from_grid(grid=np.array([[0.0, 1.0]])) == pytest.approx(0.30840, 1e-3)
        #
        # gnfw = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), kappa_s=2.0, axis_ratio=0.5,
        #                                    phi=90.0, inner_slope=1.5, scale_radius=1.0)
        # assert gnfw.convergence_from_grid(grid=np.array([[0.0, 1.0]])) == pytest.approx(0.30840 * 2, 1e-3)

    def test__potential_correct_values(self):
        gnfw = mp.SphericalGeneralizedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, inner_slope=0.5, scale_radius=8.0
        )
        assert gnfw.potential_from_grid(
            grid=np.array([[0.1625, 0.1875]])
        ) == pytest.approx(0.00920, 1e-3)

        gnfw = mp.SphericalGeneralizedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, inner_slope=1.5, scale_radius=8.0
        )
        assert gnfw.potential_from_grid(
            grid=np.array([[0.1625, 0.1875]])
        ) == pytest.approx(0.17448, 1e-3)

        # gnfw = mp.EllipticalGeneralizedNFW(centre=(1.0, 1.0), kappa_s=5.0, axis_ratio=0.5,
        #                                    phi=100.0, inner_slope=1.0, scale_radius=10.0)
        # assert gnfw.potential_from_grid(grid=np.array([[2.0, 2.0]])) == pytest.approx(2.4718, 1e-4)

    def test__deflections_correct_values(self):
        gnfw = mp.SphericalGeneralizedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, inner_slope=0.5, scale_radius=8.0
        )
        deflections = gnfw.deflections_from_grid(grid=np.array([[0.1875, 0.1625]]))
        assert deflections[0, 0] == pytest.approx(0.43501, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.37701, 1e-3)

        gnfw = mp.SphericalGeneralizedNFW(
            centre=(0.3, 0.2), kappa_s=2.5, inner_slope=1.5, scale_radius=4.0
        )
        deflections = gnfw.deflections_from_grid(grid=np.array([[0.1875, 0.1625]]))
        assert deflections[0, 0] == pytest.approx(-9.31254, 1e-3)
        assert deflections[0, 1] == pytest.approx(-3.10418, 1e-3)

        # gnfw = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), kappa_s=1.0, axis_ratio=0.3,
        #                                    phi=100.0, inner_slope=0.5, scale_radius=8.0)
        # deflections = gnfw.deflections_from_grid(grid=np.array([[0.1875, 0.1625]]))
        # assert deflections[0, 0] == pytest.approx(0.26604, 1e-3)
        # assert deflections[0, 1] == pytest.approx(0.58988, 1e-3)
        #
        # gnfw = mp.EllipticalGeneralizedNFW(centre=(0.3, 0.2), kappa_s=2.5, axis_ratio=0.5,
        #                                    phi=100.0, inner_slope=1.5, scale_radius=4.0)
        # deflections = gnfw.deflections_from_grid(grid=np.array([[0.1875, 0.1625]]))
        # assert deflections[0, 0] == pytest.approx(-5.99032, 1e-3)
        # assert deflections[0, 1] == pytest.approx(-4.02541, 1e-3)

    def test__surfce_density__change_geometry(self):
        gnfw_0 = mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))
        gnfw_1 = mp.SphericalGeneralizedNFW(centre=(1.0, 1.0))
        assert gnfw_0.convergence_from_grid(
            grid=np.array([[1.0, 1.0]])
        ) == gnfw_1.convergence_from_grid(grid=np.array([[0.0, 0.0]]))

        gnfw_0 = mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))
        gnfw_1 = mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))
        assert gnfw_0.convergence_from_grid(
            grid=np.array([[1.0, 0.0]])
        ) == gnfw_1.convergence_from_grid(grid=np.array([[0.0, 1.0]]))

        # gnfw_0 = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=0.0)
        # gnfw_1 = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0)
        # assert gnfw_0.convergence_from_grid(grid=np.array([[1.0, 0.0]])) == gnfw_1.convergence_from_grid(
        #     grid=np.array([[0.0, 1.0]]))

    def test__potential__change_geometry(self):
        gnfw_0 = mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))
        gnfw_1 = mp.SphericalGeneralizedNFW(centre=(1.0, 1.0))
        assert gnfw_0.potential_from_grid(
            grid=np.array([[1.0, 1.0]])
        ) == gnfw_1.potential_from_grid(grid=np.array([[0.0, 0.0]]))

        gnfw_0 = mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))
        gnfw_1 = mp.SphericalGeneralizedNFW(centre=(0.0, 0.0))
        assert gnfw_0.potential_from_grid(
            grid=np.array([[1.0, 0.0]])
        ) == gnfw_1.potential_from_grid(grid=np.array([[0.0, 1.0]]))

        # gnfw_0 = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=0.0)
        # gnfw_1 = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0)
        # assert gnfw_0.potential_from_grid(grid=np.array([[1.0, 0.0]])) == gnfw_1.potential_from_grid(
        #     grid=np.array([[0.0, 1.0]]))

    def test__deflections__change_geometry(self):
        gnfw_0 = mp.SphericalGeneralizedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, inner_slope=1.5, scale_radius=1.0
        )
        gnfw_1 = mp.SphericalGeneralizedNFW(
            centre=(1.0, 1.0), kappa_s=1.0, inner_slope=1.5, scale_radius=1.0
        )
        deflections_0 = gnfw_0.deflections_from_grid(grid=np.array([[1.0, 1.0]]))
        deflections_1 = gnfw_1.deflections_from_grid(grid=np.array([[0.0, 0.0]]))
        assert deflections_0[0, 0] == pytest.approx(-deflections_1[0, 0], 1e-5)
        assert deflections_0[0, 1] == pytest.approx(-deflections_1[0, 1], 1e-5)

        gnfw_0 = mp.SphericalGeneralizedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, inner_slope=1.5, scale_radius=1.0
        )
        gnfw_1 = mp.SphericalGeneralizedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, inner_slope=1.5, scale_radius=1.0
        )
        deflections_0 = gnfw_0.deflections_from_grid(grid=np.array([[1.0, 0.0]]))
        deflections_1 = gnfw_1.deflections_from_grid(grid=np.array([[0.0, 1.0]]))
        assert deflections_0[0, 0] == pytest.approx(deflections_1[0, 1], 1e-5)
        assert deflections_0[0, 1] == pytest.approx(deflections_1[0, 0], 1e-5)

        # gnfw_0 = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=0.0, kappa_s=1.0,
        #                                      inner_slope=1.5, scale_radius=1.0)
        # gnfw_1 = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, kappa_s=1.0,
        #                                      inner_slope=1.5, scale_radius=1.0)
        # deflections_0 = gnfw_0.deflections_from_grid(grid=np.array([[1.0, 0.0]]))
        # deflections_1 = gnfw_1.deflections_from_grid(grid=np.array([[0.0, 1.0]]))
        # assert deflections_0[0, 0] == pytest.approx(deflections_1[0, 1], 1e-5)
        # assert deflections_0[0, 1] == pytest.approx(deflections_1[0, 0], 1e-5)

    def test__deflections_of_spherical_profile__use_interpolate_and_cache_decorators(
        self
    ):
        gNFW = mp.SphericalGeneralizedNFW(
            centre=(-0.7, 0.5), kappa_s=1.0, inner_slope=0.5, scale_radius=8.0
        )

        mask = np.array(
            [
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, True, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ]
        )

        mask = msk.Mask(mask, pixel_scale=1.0)

        regular = grids.Grid.from_mask_and_sub_grid_size(mask=mask)

        regular_with_interp = regular.new_grid_with_interpolator(interp_pixel_scale=0.5)
        interp_deflections = gNFW.deflections_from_grid(grid=regular_with_interp)

        interpolator = grids.Interpolator.from_mask_grid_and_interp_pixel_scales(
            mask=mask, grid=regular, interp_pixel_scale=0.5
        )

        interp_deflections_values = gNFW.deflections_from_grid(
            grid=interpolator.interp_grid
        )

        interp_deflections_manual_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interp_deflections_manual_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert (interp_deflections_manual_y == interp_deflections[:, 0]).all()
        assert (interp_deflections_manual_x == interp_deflections[:, 1]).all()

    # def test__compare_to_nfw(self):
    #     nfw = mp.EllipticalNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=0.0, kappa_s=1.0, scale_radius=5.0)
    #     gnfw = mp.EllipticalGeneralizedNFW(centre=(0.0, 0.0), axis_ratio=0.8, phi=0.0, kappa_s=1.0,
    #                                        inner_slope=1.0, scale_radius=5.0)
    #
    #     assert nfw.potential_from_grid(grid) == pytest.approx(gnfw.potential_from_grid(grid), 1e-3)
    #     assert nfw.potential_from_grid(grid) == pytest.approx(gnfw.potential_from_grid(grid), 1e-3)
    #     assert nfw.deflections_from_grid(grid) == pytest.approx(gnfw.deflections_from_grid(grid), 1e-3)
    #     assert nfw.deflections_from_grid(grid) == pytest.approx(gnfw.deflections_from_grid(grid), 1e-3)

    # def test__spherical_and_elliptical_match(self):
    #     elliptical = mp.EllipticalGeneralizedNFW(centre=(0.1, 0.2), axis_ratio=1.0, phi=0.0, kappa_s=2.0,
    #                                              inner_slope=1.5, scale_radius=3.0)
    #     spherical = mp.SphericalGeneralizedNFW(centre=(0.1, 0.2), kappa_s=2.0, inner_slope=1.5, scale_radius=3.0)
    #
    #     assert elliptical.convergence_from_grid(grid) == pytest.approx(spherical.convergence_from_grid(grid),
    #                                                                        1e-4)
    #     assert elliptical.potential_from_grid(grid) == pytest.approx(spherical.potential_from_grid(grid), 1e-4)
    #     assert elliptical.deflections_from_grid(grid) == pytest.approx(spherical.deflections_from_grid(grid), 1e-4)

    def test__reshape_decorators(self):

        regular_grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(2, 2), pixel_scale=1.0
        )

        # gnfw = mp.EllipticalGeneralizedNFW()
        #
        # convergence = gnfw.convergence_from_grid(
        #     grid=regular_grid, return_in_2d=True, return_binned=False)
        #
        # assert convergence.shape == (2, 2)
        #
        # potential = gnfw.potential_from_grid(
        #     grid=regular_grid, return_in_2d=True, return_binned=False)
        #
        # assert potential.shape == (2, 2)
        #
        # deflections = gnfw.deflections_from_grid(
        #     grid=regular_grid, return_in_2d=True, return_binned=False)
        #
        # assert deflections.shape == (2, 2, 2)

        gnfw = mp.SphericalGeneralizedNFW()

        convergence = gnfw.convergence_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert convergence.shape == (2, 2)

        potential = gnfw.potential_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert potential.shape == (2, 2)

        deflections = gnfw.deflections_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert deflections.shape == (2, 2, 2)


class TestTruncatedNFW(object):
    def test__constructor_and_units(self):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(1.0, 2.0), kappa_s=2.0, scale_radius=10.0, truncation_radius=2.0
        )

        assert truncated_nfw.centre == (1.0, 2.0)
        assert isinstance(truncated_nfw.centre[0], dim.Length)
        assert isinstance(truncated_nfw.centre[1], dim.Length)
        assert truncated_nfw.centre[0].unit == "arcsec"
        assert truncated_nfw.centre[1].unit == "arcsec"

        assert truncated_nfw.axis_ratio == 1.0
        assert isinstance(truncated_nfw.axis_ratio, float)

        assert truncated_nfw.phi == 0.0
        assert isinstance(truncated_nfw.phi, float)

        assert truncated_nfw.kappa_s == 2.0
        assert isinstance(truncated_nfw.kappa_s, float)

        assert truncated_nfw.inner_slope == 1.0
        assert isinstance(truncated_nfw.inner_slope, float)

        assert truncated_nfw.scale_radius == 10.0
        assert isinstance(truncated_nfw.scale_radius, dim.Length)
        assert truncated_nfw.scale_radius.unit_length == "arcsec"

        assert truncated_nfw.truncation_radius == 2.0
        assert isinstance(truncated_nfw.truncation_radius, dim.Length)
        assert truncated_nfw.truncation_radius.unit_length == "arcsec"

    def test__convergence_correct_values(self):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0, truncation_radius=2.0
        )

        assert truncated_nfw.convergence_from_grid(
            grid=np.array([[2.0, 0.0]])
        ) == pytest.approx(2.0 * 0.046409642, 1.0e-4)

        assert truncated_nfw.convergence_from_grid(
            grid=np.array([[1.0, 1.0]])
        ) == pytest.approx(2.0 * 0.10549515, 1.0e-4)

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=3.0, scale_radius=1.0, truncation_radius=2.0
        )

        assert truncated_nfw.convergence_from_grid(
            grid=np.array([[2.0, 0.0]])
        ) == pytest.approx(6.0 * 0.046409642, 1.0e-4)

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=3.0, scale_radius=5.0, truncation_radius=2.0
        )

        assert truncated_nfw.convergence_from_grid(
            grid=np.array([[2.0, 0.0]])
        ) == pytest.approx(1.51047026, 1.0e-4)

    def test__deflections_correct_values(self):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0, truncation_radius=2.0
        )

        # factor = (4.0 * kappa_s * scale_radius / (r / scale_radius))

        deflections = truncated_nfw.deflections_from_grid(grid=np.array([[2.0, 0.0]]))

        factor = (4.0 * 1.0 * 1.0) / (2.0 / 1.0)
        assert deflections[0, 0] == pytest.approx(factor * 0.38209715, 1.0e-4)
        assert deflections[0, 1] == pytest.approx(0.0, 1.0e-4)

        deflections = truncated_nfw.deflections_from_grid(grid=np.array([[0.0, 2.0]]))

        assert deflections[0, 0] == pytest.approx(0.0, 1.0e-4)
        assert deflections[0, 1] == pytest.approx(factor * 0.38209715, 1.0e-4)

        deflections = truncated_nfw.deflections_from_grid(grid=np.array([[1.0, 1.0]]))

        factor = (4.0 * 1.0 * 1.0) / (np.sqrt(2) / 1.0)
        assert deflections[0, 0] == pytest.approx(
            (1.0 / np.sqrt(2)) * factor * 0.3125838, 1.0e-4
        )
        assert deflections[0, 1] == pytest.approx(
            (1.0 / np.sqrt(2)) * factor * 0.3125838, 1.0e-4
        )

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=2.0, scale_radius=1.0, truncation_radius=2.0
        )

        deflections = truncated_nfw.deflections_from_grid(grid=np.array([[2.0, 0.0]]))

        factor = (4.0 * 2.0 * 1.0) / (2.0 / 1.0)
        assert deflections[0, 0] == pytest.approx(factor * 0.38209715, 1.0e-4)
        assert deflections[0, 1] == pytest.approx(0.0, 1.0e-4)

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, scale_radius=4.0, truncation_radius=2.0
        )

        deflections = truncated_nfw.deflections_from_grid(grid=np.array([[2.0, 0.0]]))

        assert deflections[0, 0] == pytest.approx(2.1702661386, 1.0e-4)
        assert deflections[0, 1] == pytest.approx(0.0, 1.0e-4)

    def test__compare_nfw_and_truncated_nfw_with_large_truncation_radius__convergence_and_deflections_identical(
        self
    ):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, scale_radius=4.0, truncation_radius=50000.0
        )

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=4.0)

        truncated_nfw_convergence = truncated_nfw.convergence_from_grid(
            grid=np.array([[2.0, 2.0], [3.0, 1.0], [-1.0, -9.0]])
        )
        nfw_convergence = nfw.convergence_from_grid(
            grid=np.array([[2.0, 2.0], [3.0, 1.0], [-1.0, -9.0]])
        )

        assert truncated_nfw_convergence == pytest.approx(nfw_convergence, 1.0e-4)

        truncated_nfw_deflections = truncated_nfw.deflections_from_grid(
            grid=np.array([[2.0, 2.0], [3.0, 1.0], [-1.0, -9.0]])
        )
        nfw_deflections = nfw.deflections_from_grid(
            grid=np.array([[2.0, 2.0], [3.0, 1.0], [-1.0, -9.0]])
        )

        assert truncated_nfw_deflections == pytest.approx(nfw_deflections, 1.0e-4)

    def test__deflections_of_spherical_profile__dont_use_interpolate_and_cache_decorators(
        self
    ):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(-0.7, 0.5), kappa_s=1.0, scale_radius=8.0, truncation_radius=2.0
        )

        mask = np.array(
            [
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, True, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ]
        )

        mask = msk.Mask(mask, pixel_scale=1.0)

        regular = grids.Grid.from_mask_and_sub_grid_size(mask=mask)

        regular_with_interp = regular.new_grid_with_interpolator(interp_pixel_scale=0.5)
        interp_deflections = truncated_nfw.deflections_from_grid(
            grid=regular_with_interp
        )

        interpolator = grids.Interpolator.from_mask_grid_and_interp_pixel_scales(
            mask=mask, grid=regular, interp_pixel_scale=0.5
        )

        interp_deflections_values = truncated_nfw.deflections_from_grid(
            grid=interpolator.interp_grid
        )

        interp_deflections_manual_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interp_deflections_manual_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert (interp_deflections_manual_y != interp_deflections[:, 0]).all()
        assert (interp_deflections_manual_x != interp_deflections[:, 1]).all()

    def test__mass_at_truncation_radius__values(self):

        truncated_nfw = mp.SphericalTruncatedNFW(
            centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0, truncation_radius=1.0
        )

        cosmology = mock_cosmology.MockCosmology(
            arcsec_per_kpc=1.0,
            kpc_per_arcsec=1.0,
            critical_surface_density=1.0,
            cosmic_average_density=1.0,
        )

        mass_at_truncation_radius = truncated_nfw.mass_at_truncation_radius(
            redshift_profile=0.5,
            redshift_source=1.0,
            unit_length="arcsec",
            unit_mass="solMass",
            cosmology=cosmology,
        )

        assert mass_at_truncation_radius == pytest.approx(0.00009792581, 1.0e-5)

        # truncated_nfw = mp.SphericalTruncatedNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0,
        #                                          truncation_radius=1.0)
        #
        # cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=1.0, kpc_per_arcsec=1.0, critical_surface_density=2.0,
        #                           cosmic_average_density=3.0)
        #
        # mass_at_truncation_radius = truncated_nfw.mass_at_truncation_radius(redshift_lens=0.5, redshift_source=1.0,
        #     unit_length='arcsec', unit_mass='solMass', cosmology=cosmology)
        #
        # assert mass_at_truncation_radius == pytest.approx(0.00008789978, 1.0e-5)
        #
        # truncated_nfw = mp.SphericalTruncatedNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=2.0,
        #                                          truncation_radius=1.0)
        #
        # mass_at_truncation_radius = truncated_nfw.mass_at_truncation_radius(redshift_lens=0.5, redshift_source=1.0,
        #     unit_length='arcsec', unit_mass='solMass', cosmology=cosmology)
        #
        # assert mass_at_truncation_radius == pytest.approx(0.0000418378, 1.0e-5)
        #
        # truncated_nfw = mp.SphericalTruncatedNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=8.0,
        #                                          truncation_radius=4.0)
        #
        # mass_at_truncation_radius = truncated_nfw.mass_at_truncation_radius(redshift_lens=0.5, redshift_source=1.0,
        #     unit_length='arcsec', unit_mass='solMass', cosmology=cosmology)
        #
        # assert mass_at_truncation_radius == pytest.approx(0.0000421512, 1.0e-4)
        #
        # truncated_nfw = mp.SphericalTruncatedNFW(centre=(0.0, 0.0), kappa_s=2.0, scale_radius=8.0,
        #                                          truncation_radius=4.0)
        #
        # mass_at_truncation_radius = truncated_nfw.mass_at_truncation_radius(redshift_lens=0.5, redshift_source=1.0,
        #     unit_length='arcsec', unit_mass='solMass', cosmology=cosmology)
        #
        # assert mass_at_truncation_radius == pytest.approx(0.00033636625, 1.0e-4)

    def test_summarize_in_units(self):

        cosmology = cosmo.LambdaCDM(H0=70.0, Om0=0.3, Ode0=0.7)

        nfw = mp.SphericalTruncatedNFW(
            kappa_s=0.5, scale_radius=5.0, truncation_radius=10.0
        )

        summary_text = nfw.summarize_in_units(
            radii=[dim.Length(10.0), dim.Length(500.0)],
            prefix="nfw_",
            unit_length="kpc",
            unit_mass="solMass",
            redshift_profile=0.6,
            redshift_source=2.5,
            redshift_of_cosmic_average_density="profile",
            whitespace=50,
            cosmology=cosmology,
        )

        i = 0

        assert summary_text[i] == "Mass Profile = SphericalTruncatedNFW\n"
        i += 1
        assert (
            summary_text[i]
            == "nfw_einstein_radius                               15.38 kpc"
        )
        i += 1
        assert (
            summary_text[i]
            == "nfw_einstein_mass                                 1.4418e+12 solMass"
        )
        i += 1
        assert (
            summary_text[i]
            == "nfw_mass_within_10.00_kpc                         5.2061e+12 solMass"
        )
        i += 1
        assert (
            summary_text[i]
            == "nfw_mass_within_500.00_kpc                        7.3287e+12 solMass"
        )
        i += 1
        assert (
            summary_text[i]
            == "nfw_rho_at_scale_radius                           29027857.02 solMass/kpc3"
        )
        i += 1
        assert (
            summary_text[i]
            == "nfw_delta_concentration                           110665.28"
        )
        i += 1
        assert (
            summary_text[i] == "nfw_concentration                                 14.40"
        )
        i += 1
        assert (
            summary_text[i]
            == "nfw_radius_at_200x_cosmic_density                 481.41 kpc"
        )
        i += 1
        assert (
            summary_text[i]
            == "nfw_mass_at_200x_cosmic_density                   2.4517e+13 solMass"
        )
        i += 1
        assert (
            summary_text[i]
            == "nfw_mass_at_truncation_radius                     1.3190e+13 solMass"
        )
        i += 1

    def test__reshape_decorators(self):

        regular_grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(2, 2), pixel_scale=1.0
        )

        truncated_nfw = mp.SphericalTruncatedNFW()

        convergence = truncated_nfw.convergence_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert convergence.shape == (2, 2)

        potential = truncated_nfw.potential_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert potential.shape == (2, 2)

        deflections = truncated_nfw.deflections_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert deflections.shape == (2, 2, 2)


class TestTruncatedNFWChallenge(object):
    def test__constructor_and_units(self):

        truncated_nfw = mp.SphericalTruncatedNFWChallenge(
            centre=(1.0, 2.0), kappa_s=2.0, scale_radius=10.0
        )

        assert truncated_nfw.centre == (1.0, 2.0)
        assert isinstance(truncated_nfw.centre[0], dim.Length)
        assert isinstance(truncated_nfw.centre[1], dim.Length)
        assert truncated_nfw.centre[0].unit == "arcsec"
        assert truncated_nfw.centre[1].unit == "arcsec"

        assert truncated_nfw.axis_ratio == 1.0
        assert isinstance(truncated_nfw.axis_ratio, float)

        assert truncated_nfw.phi == 0.0
        assert isinstance(truncated_nfw.phi, float)

        assert truncated_nfw.kappa_s == 2.0
        assert isinstance(truncated_nfw.kappa_s, float)

        assert truncated_nfw.inner_slope == 1.0
        assert isinstance(truncated_nfw.inner_slope, float)

        assert truncated_nfw.scale_radius == 10.0
        assert isinstance(truncated_nfw.scale_radius, dim.Length)
        assert truncated_nfw.scale_radius.unit_length == "arcsec"

        assert truncated_nfw.truncation_radius == pytest.approx(
            2.0 * 189.26967095554755, 1.0e-4
        )
        assert isinstance(truncated_nfw.truncation_radius, dim.Length)
        assert truncated_nfw.truncation_radius.unit_length == "arcsec"


class TestNFW(object):
    def test__constructor_and_units(self):

        nfw = mp.EllipticalNFW(
            centre=(1.0, 2.0), axis_ratio=0.5, phi=45.0, kappa_s=2.0, scale_radius=10.0
        )

        assert nfw.centre == (1.0, 2.0)
        assert isinstance(nfw.centre[0], dim.Length)
        assert isinstance(nfw.centre[1], dim.Length)
        assert nfw.centre[0].unit == "arcsec"
        assert nfw.centre[1].unit == "arcsec"

        assert nfw.axis_ratio == 0.5
        assert isinstance(nfw.axis_ratio, float)

        assert nfw.phi == 45.0
        assert isinstance(nfw.phi, float)

        assert nfw.kappa_s == 2.0
        assert isinstance(nfw.kappa_s, float)

        assert nfw.inner_slope == 1.0
        assert isinstance(nfw.inner_slope, float)

        assert nfw.scale_radius == 10.0
        assert isinstance(nfw.scale_radius, dim.Length)
        assert nfw.scale_radius.unit_length == "arcsec"

        nfw = mp.SphericalNFW(centre=(1.0, 2.0), kappa_s=2.0, scale_radius=10.0)

        assert nfw.centre == (1.0, 2.0)
        assert isinstance(nfw.centre[0], dim.Length)
        assert isinstance(nfw.centre[1], dim.Length)
        assert nfw.centre[0].unit == "arcsec"
        assert nfw.centre[1].unit == "arcsec"

        assert nfw.axis_ratio == 1.0
        assert isinstance(nfw.axis_ratio, float)

        assert nfw.phi == 0.0
        assert isinstance(nfw.phi, float)

        assert nfw.kappa_s == 2.0
        assert isinstance(nfw.kappa_s, float)

        assert nfw.inner_slope == 1.0
        assert isinstance(nfw.inner_slope, float)

        assert nfw.scale_radius == 10.0
        assert isinstance(nfw.scale_radius, dim.Length)
        assert nfw.scale_radius.unit_length == "arcsec"

    def test__convergence_correct_values(self):
        # r = 2.0 (> 1.0)
        # F(r) = (1/(sqrt(3))*atan(sqrt(3)) = 0.60459978807
        # kappa(r) = 2 * kappa_s * (1 - 0.60459978807) / (4-1) = 0.263600141
        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)
        assert nfw.convergence_from_grid(grid=np.array([[2.0, 0.0]])) == pytest.approx(
            0.263600141, 1e-3
        )

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)
        assert nfw.convergence_from_grid(grid=np.array([[0.5, 0.0]])) == pytest.approx(
            1.388511, 1e-3
        )

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=2.0, scale_radius=1.0)
        assert nfw.convergence_from_grid(grid=np.array([[0.5, 0.0]])) == pytest.approx(
            2.0 * 1.388511, 1e-3
        )

        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=2.0)
        assert nfw.convergence_from_grid(grid=np.array([[1.0, 0.0]])) == pytest.approx(
            1.388511, 1e-3
        )

        nfw = mp.EllipticalNFW(
            centre=(0.0, 0.0), axis_ratio=0.5, phi=0.0, kappa_s=1.0, scale_radius=1.0
        )
        assert nfw.convergence_from_grid(grid=np.array([[0.25, 0.0]])) == pytest.approx(
            1.388511, 1e-3
        )

    def test__potential_correct_values(self):
        nfw = mp.SphericalNFW(centre=(0.3, 0.2), kappa_s=2.5, scale_radius=4.0)
        assert nfw.potential_from_grid(
            grid=np.array([[0.1875, 0.1625]])
        ) == pytest.approx(0.03702, 1e-3)

        nfw = mp.SphericalNFW(centre=(0.3, 0.2), kappa_s=2.5, scale_radius=4.0)
        assert nfw.potential_from_grid(
            grid=np.array([[0.1875, 0.1625]])
        ) == pytest.approx(0.03702, 1e-3)

        nfw = mp.EllipticalNFW(
            centre=(0.3, 0.2), axis_ratio=0.7, phi=6.0, kappa_s=2.5, scale_radius=4.0
        )
        assert nfw.potential_from_grid(
            grid=np.array([[0.1625, 0.1625]])
        ) == pytest.approx(0.05380, 1e-3)

    def test__potential__spherical_and_elliptical_are_same(self):
        nfw_spherical = mp.SphericalNFW(
            centre=(0.3, 0.2), kappa_s=2.5, scale_radius=4.0
        )
        nfw_elliptical = mp.EllipticalNFW(
            centre=(0.3, 0.2), axis_ratio=1.0, phi=0.0, kappa_s=2.5, scale_radius=4.0
        )

        potential_spherical = nfw_spherical.potential_from_grid(
            grid=np.array([[0.1875, 0.1625]])
        )
        potential_elliptical = nfw_elliptical.potential_from_grid(
            grid=np.array([[0.1875, 0.1625]])
        )

        assert potential_spherical == pytest.approx(potential_elliptical, 1e-3)

        potential_spherical = nfw_spherical.potential_from_grid(
            grid=np.array([[50.0, 50.0]])
        )
        potential_elliptical = nfw_elliptical.potential_from_grid(
            grid=np.array([[50.0, 50.0]])
        )

        assert potential_spherical == pytest.approx(potential_elliptical, 1e-3)

        potential_spherical = nfw_spherical.potential_from_grid(
            grid=np.array([[-50.0, -50.0]])
        )
        potential_elliptical = nfw_elliptical.potential_from_grid(
            grid=np.array([[-50.0, -50.0]])
        )

        assert potential_spherical == pytest.approx(potential_elliptical, 1e-3)

    def test__deflections_correct_values(self):
        nfw = mp.SphericalNFW(centre=(0.0, 0.0), kappa_s=1.0, scale_radius=1.0)
        deflections = nfw.deflections_from_grid(grid=np.array([[0.1625, 0.1625]]))
        assert deflections[0, 0] == pytest.approx(0.56194, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.56194, 1e-3)

        nfw = mp.SphericalNFW(centre=(0.3, 0.2), kappa_s=2.5, scale_radius=4.0)
        deflections = nfw.deflections_from_grid(grid=np.array([[0.1875, 0.1625]]))
        assert deflections[0, 0] == pytest.approx(-2.08909, 1e-3)
        assert deflections[0, 1] == pytest.approx(-0.69636, 1e-3)

        nfw = mp.EllipticalNFW(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, kappa_s=1.0, scale_radius=1.0
        )
        deflections = nfw.deflections_from_grid(grid=np.array([[0.1625, 0.1625]]))
        assert deflections[0, 0] == pytest.approx(0.56194, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.56194, 1e-3)

        nfw = mp.EllipticalNFW(
            centre=(0.3, 0.2), axis_ratio=0.7, phi=6.0, kappa_s=2.5, scale_radius=4.0
        )
        deflections = nfw.deflections_from_grid(grid=np.array([[0.1625, 0.1625]]))
        assert deflections[0, 0] == pytest.approx(-2.59480, 1e-3)
        assert deflections[0, 1] == pytest.approx(-0.44204, 1e-3)

    def test__deflections_of_elliptical_profile__use_interpolate_and_cache_decorators(
        self
    ):
        nfw = mp.EllipticalNFW(
            centre=(-0.7, 0.5), axis_ratio=0.9, phi=45.0, kappa_s=1.0, scale_radius=8.0
        )

        mask = np.array(
            [
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, True, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ]
        )

        mask = msk.Mask(mask, pixel_scale=1.0)

        regular = grids.Grid.from_mask_and_sub_grid_size(mask=mask)

        regular_with_interp = regular.new_grid_with_interpolator(interp_pixel_scale=0.5)
        interp_deflections = nfw.deflections_from_grid(grid=regular_with_interp)

        interpolator = grids.Interpolator.from_mask_grid_and_interp_pixel_scales(
            mask=mask, grid=regular, interp_pixel_scale=0.5
        )

        interp_deflections_values = nfw.deflections_from_grid(
            grid=interpolator.interp_grid
        )

        interp_deflections_manual_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interp_deflections_manual_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert (interp_deflections_manual_y == interp_deflections[:, 0]).all()
        assert (interp_deflections_manual_x == interp_deflections[:, 1]).all()

    def test__deflections_of_spherical_profile__dont_use_interpolate_and_cache_decorators(
        self
    ):
        nfw = mp.SphericalNFW(centre=(-0.7, 0.5), kappa_s=1.0, scale_radius=8.0)

        mask = np.array(
            [
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, True, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ]
        )

        mask = msk.Mask(mask, pixel_scale=1.0)

        regular = grids.Grid.from_mask_and_sub_grid_size(mask=mask)

        regular_with_interp = regular.new_grid_with_interpolator(interp_pixel_scale=0.5)
        interp_deflections = nfw.deflections_from_grid(grid=regular_with_interp)

        interpolator = grids.Interpolator.from_mask_grid_and_interp_pixel_scales(
            mask=mask, grid=regular, interp_pixel_scale=0.5
        )

        interp_deflections_values = nfw.deflections_from_grid(
            grid=interpolator.interp_grid
        )

        interp_deflections_manual_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interp_deflections_manual_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert (interp_deflections_manual_y != interp_deflections[:, 0]).all()
        assert (interp_deflections_manual_x != interp_deflections[:, 1]).all()

    def test__reshape_decorators(self):

        regular_grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=(2, 2), pixel_scale=1.0
        )

        nfw = mp.EllipticalNFW()

        convergence = nfw.convergence_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert convergence.shape == (2, 2)

        potential = nfw.potential_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert potential.shape == (2, 2)

        deflections = nfw.deflections_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert deflections.shape == (2, 2, 2)

        nfw = mp.SphericalNFW()

        convergence = nfw.convergence_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert convergence.shape == (2, 2)

        potential = nfw.potential_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert potential.shape == (2, 2)

        deflections = nfw.deflections_from_grid(
            grid=regular_grid, return_in_2d=True, return_binned=False
        )

        assert deflections.shape == (2, 2, 2)



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
from matplotlib import pyplot as plt

from autolens.plotters import plotter_util, array_plotters
from autolens.model.profiles.plotters import profile_plotters


def plot_intensities(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Intensities",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_intensities",
):
    """Plot the intensities (e.g. the datas) of a galaxy, on a regular grid of (y,x) coordinates.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.galaxy.Galaxy
        The galaxy whose intensities are plotted.
    grid : ndarray or datas.array.grid_stacks.RegularGrid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
    """
    intensities = galaxy.intensities_from_grid(
        grid=grid, return_in_2d=True, return_binned=True
    )

    array_plotters.plot_array(
        array=intensities,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_convergence(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_critical_curves=True,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Convergence",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_convergence",
):
    """Plot the convergence of a galaxy, on a regular grid of (y,x) coordinates.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.galaxy.Galaxy
        The galaxy whose convergence is plotted.
    grid : ndarray or datas.array.grid_stacks.RegularGrid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
    """
    convergence = galaxy.convergence_from_grid(
        grid=grid, return_in_2d=True, return_binned=True
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=galaxy,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=convergence,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_potential(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_critical_curves=True,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Potential",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_potential",
):
    """Plot the potential of a galaxy, on a regular grid of (y,x) coordinates.

     Set *autolens.datas.array.plotters.array_plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
    galaxy : model.galaxy.galaxy.Galaxy
         The galaxy whose potential is plotted.
    grid : ndarray or datas.array.grid_stacks.RegularGrid
         The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
     """
    potential = galaxy.potential_from_grid(
        grid=grid, return_in_2d=True, return_binned=True
    )

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=galaxy,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=potential,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_deflections_y(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_critical_curves=True,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Deflections (y)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_deflections_y",
):
    """Plot the y component of the deflection angles of a galaxy, on a regular grid of (y,x) coordinates.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.galaxy.Galaxy
        The galaxy whose y deflecton angles are plotted.
    grid : ndarray or datas.array.grid_stacks.RegularGrid
        The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
    """
    deflections = galaxy.deflections_from_grid(grid)
    deflections_y = grid.scaled_array_2d_from_array_1d(deflections[:, 0])

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=galaxy,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=deflections_y,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_deflections_x(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_critical_curves=True,
    plot_caustics=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Deflections (x)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_deflections_x",
):
    """Plot the x component of the deflection angles of a galaxy, on a regular grid of (y,x) coordinates.

     Set *autolens.datas.array.plotters.array_plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
    galaxy : model.galaxy.galaxy.Galaxy
         The galaxy whose x deflecton angles are plotted.
     grid : ndarray or datas.array.grid_stacks.RegularGrid
         The (y,x) coordinates of the grid, in an array of shape (total_coordinates, 2)
     """
    deflections = galaxy.deflections_from_grid(grid)
    deflections_x = grid.scaled_array_2d_from_array_1d(deflections[:, 1])

    lines = plotter_util.get_critical_curve_and_caustic(
        obj=galaxy,
        grid=grid,
        plot_critical_curve=plot_critical_curves,
        plot_caustics=plot_caustics,
    )

    array_plotters.plot_array(
        array=deflections_x,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        lines=lines,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_intensities_subplot(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_intensities",
):

    total_light_profiles = len(galaxy.light_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_light_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, light_profile in enumerate(galaxy.light_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.plot_intensities(
            light_profile=light_profile,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            grid=grid,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def plot_convergence_subplot(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_convergence",
):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_mass_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.plot_convergence(
            mass_profile=mass_profile,
            grid=grid,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def plot_potential_subplot(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_potential",
):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_mass_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.plot_potential(
            mass_profile=mass_profile,
            grid=grid,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def plot_deflections_y_subplot(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_deflections_y",
):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_mass_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.plot_deflections_y(
            mass_profile=mass_profile,
            grid=grid,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def plot_deflections_x_subplot(
    galaxy,
    grid,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_deflections_x",
):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_mass_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.plot_deflections_x(
            mass_profile=mass_profile,
            grid=grid,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            as_subplot=True,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()



import numpy as np

import autofit as af
from autolens import exc
from autolens.lens.util import lens_fit_util as util
from autolens.model.galaxy import galaxy as g

from autolens.data.array.grids import reshape_returned_regular_array


class LensDataFit(af.DataFit1D):
    def __init__(self, image_1d, noise_map_1d, mask_1d, model_image_1d, grid_stack):

        super().__init__(
            data_1d=image_1d,
            noise_map_1d=noise_map_1d,
            mask_1d=mask_1d,
            model_data_1d=model_image_1d,
        )

        self.grid_stack = grid_stack

    @property
    def image_1d(self):
        return self.data_1d

    @property
    def model_image_1d(self):
        return self.model_data_1d

    def mask(self, return_in_2d=True):
        if return_in_2d:
            return self.mask_2d
        else:
            return self.mask_1d

    @reshape_returned_regular_array
    def image(self):
        return self.image_1d

    @reshape_returned_regular_array
    def noise_map(self):
        return self.noise_map_1d

    @reshape_returned_regular_array
    def signal_to_noise_map(self):
        return self.signal_to_noise_map_1d

    @reshape_returned_regular_array
    def model_image(self):
        return self.model_image_1d

    @reshape_returned_regular_array
    def residual_map(self):
        return self.residual_map_1d

    @reshape_returned_regular_array
    def normalized_residual_map(self):
        return self.normalized_residual_map_1d

    @reshape_returned_regular_array
    def chi_squared_map(self):
        return self.chi_squared_map_1d

    @property
    def figure_of_merit(self):
        return self.likelihood

    @classmethod
    def for_data_and_tracer(
        cls, lens_data, tracer, hyper_image_sky=None, hyper_background_noise=None
    ):
        """Fit lens instrument with a model tracer, automatically determining the type of fit based on the \
        properties of the galaxies in the tracer.

        Parameters
        -----------
        lens_data : lens_data.LensData or lens_data.LensDataHyper
            The lens-images that is fitted.
        tracer : ray_tracing.TracerNonStack
            The tracer, which describes the ray-tracing and strong lens configuration.
        """

        if tracer.has_light_profile and not tracer.has_pixelization:
            return LensProfileFit(
                lens_data=lens_data,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )
        elif not tracer.has_light_profile and tracer.has_pixelization:
            return LensInversionFit(
                lens_data=lens_data,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )
        elif tracer.has_light_profile and tracer.has_pixelization:
            return LensProfileInversionFit(
                lens_data=lens_data,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )
        else:
            raise exc.FittingException(
                "The fit routine did not call a Fit class - check the "
                "properties of the tracer"
            )


class LensTracerFit(LensDataFit):
    def __init__(
        self,
        image_1d,
        noise_map_1d,
        mask_1d,
        model_image_1d,
        tracer,
        psf,
        mask_2d,
        positions,
    ):
        """ An  lens fitter, which contains the tracer's used to perform the fit and functions to manipulate \
        the lens instrument's hyper_galaxy.

        Parameters
        -----------
        tracer : ray_tracing.Tracer
            The tracer, which describes the ray-tracing and strong lens configuration.
        scaled_array_2d_from_array_1d : func
            A function which maps the 1D lens hyper_galaxy to its unmasked 2D array.
        """

        self.mask_2d = mask_2d
        self.positions = positions

        super().__init__(
            image_1d=image_1d,
            noise_map_1d=noise_map_1d,
            mask_1d=mask_1d,
            model_image_1d=model_image_1d,
            grid_stack=tracer.grid_stack,
        )

        self.tracer = tracer
        self.psf = psf

    @property
    def galaxy_image_1d_dict(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """
        raise NotImplementedError()

    @property
    def galaxy_image_2d_dict(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_image_2d_dict = {}

        for galalxy, galaxy_image in self.galaxy_image_1d_dict.items():

            galaxy_image_2d_dict[
                galalxy
            ] = self.grid_stack.regular.scaled_array_2d_from_array_1d(
                array_1d=galaxy_image
            )

        return galaxy_image_2d_dict

    @property
    def total_inversions(self):
        return len(self.tracer.mappers_of_planes)

    @property
    def unmasked_blurred_profile_image_plane_image(self):
        return self.tracer.unmasked_blurred_profile_image_plane_image_from_psf(
            psf=self.psf
        )

    @property
    def unmasked_blurred_profile_image_plane_image_of_planes(self):
        return self.tracer.unmasked_blurred_profile_image_plane_image_of_planes_from_psf(
            psf=self.psf
        )

    @property
    def unmasked_blurred_profile_image_plane_image_of_planes_and_galaxies(self):
        return self.tracer.unmasked_blurred_profile_image_plane_image_of_plane_and_galaxies_from_psf(
            psf=self.psf
        )


class LensProfileFit(LensTracerFit):
    def __init__(
        self, lens_data, tracer, hyper_image_sky=None, hyper_background_noise=None
    ):
        """ An  lens profile fitter, which generates the image-plane image of all galaxies (with light \
        profiles) in the tracer and blurs it with the lens instrument's PSF.

        If a padded tracer is supplied, the blurred profile image's can be generated over the entire image and thus \
        without the mask.

        Parameters
        -----------
        lens_data : lens_data.LensData
            The lens-image that is fitted.
        tracer : ray_tracing.AbstractTracerData
            The tracer, which describes the ray-tracing and strong lens configuration.
        """

        image_1d = image_1d_from_lens_data_and_hyper_image_sky(
            lens_data=lens_data, hyper_image_sky=hyper_image_sky
        )

        noise_map_1d = noise_map_1d_from_lens_data_tracer_and_hyper_backkground_noise(
            lens_data=lens_data,
            tracer=tracer,
            hyper_background_noise=hyper_background_noise,
        )

        blurred_profile_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
            convolver_image=lens_data.convolver_image
        )

        super(LensProfileFit, self).__init__(
            image_1d=image_1d,
            noise_map_1d=noise_map_1d,
            mask_1d=lens_data.mask_1d,
            model_image_1d=blurred_profile_image_1d,
            mask_2d=lens_data.mask_2d,
            tracer=tracer,
            psf=lens_data.psf,
            positions=lens_data.positions,
        )

        self.convolver_image = lens_data.convolver_image

    @property
    def galaxy_image_1d_dict(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """
        return self.tracer.galaxy_image_dict_from_convolver_image(
            convolver_image=self.convolver_image
        )

    @property
    def model_image_2d_of_planes(self):
        return self.tracer.blurred_profile_image_plane_image_2d_of_planes_from_convolver_image(
            convolver_image=self.convolver_image
        )

    @property
    def figure_of_merit(self):
        return self.likelihood


class InversionFit(LensTracerFit):
    def __init__(
        self, lens_data, image_1d, noise_map_1d, model_image_1d, tracer, inversion
    ):

        super().__init__(
            image_1d=image_1d,
            noise_map_1d=noise_map_1d,
            mask_1d=lens_data.mask_1d,
            model_image_1d=model_image_1d,
            mask_2d=lens_data.mask_2d,
            psf=lens_data.psf,
            tracer=tracer,
            positions=lens_data.positions,
        )

        self.inversion = inversion

        self.likelihood_with_regularization = util.likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
            chi_squared=self.chi_squared,
            regularization_term=inversion.regularization_term,
            noise_normalization=self.noise_normalization,
        )

        self.evidence = util.evidence_from_inversion_terms(
            chi_squared=self.chi_squared,
            regularization_term=inversion.regularization_term,
            log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
            log_regularization_term=inversion.log_det_regularization_matrix_term,
            noise_normalization=self.noise_normalization,
        )


class LensInversionFit(InversionFit):
    def __init__(
        self, lens_data, tracer, hyper_image_sky=None, hyper_background_noise=None
    ):
        """ An  lens inversion fitter, which fits the lens instrument an inversion using the mapper(s) and \
        regularization(s) in the galaxies of the tracer.

        This inversion use's the lens-image, its PSF and an input noise-map.

        Parameters
        -----------
        lens_data : lens_data.LensData
            The lens-image that is fitted.
        tracer : ray_tracing.AbstractTracerData
            The tracer, which describes the ray-tracing and strong lens configuration.
        """

        image_1d = image_1d_from_lens_data_and_hyper_image_sky(
            lens_data=lens_data, hyper_image_sky=hyper_image_sky
        )

        noise_map_1d = noise_map_1d_from_lens_data_tracer_and_hyper_backkground_noise(
            lens_data=lens_data,
            tracer=tracer,
            hyper_background_noise=hyper_background_noise,
        )

        inversion = tracer.inversion_from_image_1d_noise_map_1d_and_convolver_mapping_matrix(
            image_1d=image_1d,
            noise_map_1d=noise_map_1d,
            convolver_mapping_matrix=lens_data.convolver_mapping_matrix,
        )

        super().__init__(
            lens_data=lens_data,
            image_1d=image_1d,
            noise_map_1d=noise_map_1d,
            model_image_1d=inversion.reconstructed_data_1d,
            tracer=tracer,
            inversion=inversion,
        )

    @property
    def galaxy_image_1d_dict(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """
        galaxy_image_dict = self.tracer.galaxy_image_dict_blank_images
        galaxy_image_dict.update(
            {self.tracer.planes[-1].galaxies[0]: self.inversion.reconstructed_data_1d}
        )
        return galaxy_image_dict

    @property
    def figure_of_merit(self):
        return self.evidence

    @property
    def model_image_2d_of_planes(self):
        return [
            np.zeros(shape=self.inversion.reconstructed_data_2d.shape),
            self.inversion.reconstructed_data_2d,
        ]


class LensProfileInversionFit(InversionFit):
    def __init__(
        self, lens_data, tracer, hyper_image_sky=None, hyper_background_noise=None
    ):
        """ An  lens profile and inversion fitter, which first generates and subtracts the image-plane \
        image of all galaxies (with light profiles) in the tracer, blurs it with the PSF and fits the residual image \
        with an inversion using the mapper(s) and regularization(s) in the galaxy's of the tracer.

        This inversion use's the lens-image, its PSF and an input noise-map.

        If a padded tracer is supplied, the blurred profile image's can be generated over the entire image and thus \
        without the mask.

        Parameters
        -----------
        lens_data : lens_data.LensData
            The lens-image that is fitted.
        tracer : ray_tracing.Tracer
            The tracer, which describes the ray-tracing and strong lens configuration.
        """

        image_1d = image_1d_from_lens_data_and_hyper_image_sky(
            lens_data=lens_data, hyper_image_sky=hyper_image_sky
        )

        noise_map_1d = noise_map_1d_from_lens_data_tracer_and_hyper_backkground_noise(
            lens_data=lens_data,
            tracer=tracer,
            hyper_background_noise=hyper_background_noise,
        )

        self.blurred_profile_image_1d = tracer.blurred_profile_image_plane_image_1d_from_convolver_image(
            convolver_image=lens_data.convolver_image
        )

        self.profile_subtracted_image_1d = image_1d - self.blurred_profile_image_1d

        inversion = tracer.inversion_from_image_1d_noise_map_1d_and_convolver_mapping_matrix(
            image_1d=self.profile_subtracted_image_1d,
            noise_map_1d=noise_map_1d,
            convolver_mapping_matrix=lens_data.convolver_mapping_matrix,
        )

        model_image = self.blurred_profile_image_1d + inversion.reconstructed_data_1d

        super(LensProfileInversionFit, self).__init__(
            tracer=tracer,
            lens_data=lens_data,
            image_1d=image_1d,
            noise_map_1d=noise_map_1d,
            model_image_1d=model_image,
            inversion=inversion,
        )

        self.convolver_image = lens_data.convolver_image

    @property
    def galaxy_image_1d_dict(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """
        galaxy_image_dict = self.tracer.galaxy_image_dict_from_convolver_image(
            convolver_image=self.convolver_image
        )
        galaxy_image_dict.update(
            {self.tracer.planes[-1].galaxies[0]: self.inversion.reconstructed_data_1d}
        )
        return galaxy_image_dict

    @reshape_returned_regular_array
    def blurred_profile_image(self, return_in_2d=True):
        return self.blurred_profile_image_1d

    @reshape_returned_regular_array
    def profile_subtracted_image(self, return_in_2d=True):
        return self.profile_subtracted_image_1d

    @property
    def model_image_2d_of_planes(self):
        return [
            self.blurred_profile_image(return_in_2d=True),
            self.inversion.reconstructed_data_2d,
        ]

    @property
    def figure_of_merit(self):
        return self.evidence


class LensPositionFit(object):
    def __init__(self, positions, noise_map):
        """A lens position fitter, which takes a set of positions (e.g. from a plane in the tracer) and computes \
        their maximum separation, such that points which tracer closer to one another have a higher likelihood.

        Parameters
        -----------
        positions : [[]]
            The (y,x) arc-second coordinates of positions which the maximum distance and likelihood is computed using.
        noise_map : ndarray | float
            The noise-value assumed when computing the likelihood.
        """
        self.positions = positions
        self.noise_map = noise_map

    @property
    def chi_squared_map(self):
        return np.square(np.divide(self.maximum_separations, self.noise_map))

    @property
    def figure_of_merit(self):
        return -0.5 * sum(self.chi_squared_map)

    def maximum_separation_within_threshold(self, threshold):
        return max(self.maximum_separations) <= threshold

    @property
    def maximum_separations(self):
        return list(
            map(
                lambda positions: self.max_separation_of_grid(positions), self.positions
            )
        )

    @staticmethod
    def max_separation_of_grid(grid):
        rdist_max = np.zeros((grid.shape[0]))
        for i in range(grid.shape[0]):
            xdists = np.square(np.subtract(grid[i, 0], grid[:, 0]))
            ydists = np.square(np.subtract(grid[i, 1], grid[:, 1]))
            rdist_max[i] = np.max(np.add(xdists, ydists))
        return np.max(np.sqrt(rdist_max))


def image_1d_from_lens_data_and_hyper_image_sky(lens_data, hyper_image_sky):

    if hyper_image_sky is not None:
        return hyper_image_sky.image_scaled_sky_from_image(image=lens_data.image_1d)
    else:
        return lens_data.image_1d


def noise_map_1d_from_lens_data_tracer_and_hyper_backkground_noise(
    lens_data, tracer, hyper_background_noise
):

    if hyper_background_noise is not None:
        noise_map_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
            noise_map=lens_data.noise_map_1d
        )
    else:
        noise_map_1d = lens_data.noise_map_1d

    hyper_noise_map_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
        noise_map_1d=lens_data.noise_map_1d
    )

    if hyper_noise_map_1d is not None:
        noise_map_1d = noise_map_1d + hyper_noise_map_1d
        if lens_data.hyper_noise_map_max is not None:
            noise_map_1d[
                noise_map_1d > lens_data.hyper_noise_map_max
            ] = lens_data.hyper_noise_map_max

    return noise_map_1d



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline
from autolens.pipeline import pipeline_tagging
from autolens.model.profiles import mass_profiles as mp
from autolens.model.profiles import light_and_mass_profiles as lmp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os


# In this pipeline, we'll perform an analysis which fits an image with the lens light included, and a source galaxy
# using an inversion, using a decomposed light and dark matter profile. The pipeline follows on from the
# inversion pipeline 'pipelines/with_lens_light/inversion/from_initialize/lens_bulge_disk_sie__source_inversion.py'.

# Alignment of the centre, phi and axis-ratio of the light profile's EllipticalSersic and EllipticalExponential
# profiles use the alignment specified in the previous pipeline.

# The pipeline is two phases, as follows:

# Phase 1:

# Description: Fits the lens light and mass model as a decomposed profile, using an inversion for the Source.
# Lens Light & Mass: EllipticalSersic + EllipticalExponential
# Lens Mass: SphericalNFW + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: with_lens_light/bulge_disk/inversion/from_initialize/lens_bulge_disk_sieexp_source_inversion.py
# Prior Passing: Lens Light (variable -> previous pipeline), Lens Mass (default),
#                Source Inversion (variable / constant -> previous pipeline)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations by default.

# Phase 2:

# Description: Refines the inversion parameters, using a fixed mass model from phase 1.
# Lens Light & Mass: EllipticalSersic
# Lens Mass: SphericalNFW + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: None
# Prior Passing: Lens Light & Mass (constant -> phase 1), source inversion (variable -> phase 1)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations by default.


def make_pipeline(
    pipeline_settings,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=0.05,
    use_inversion_border=True,
    inversion_pixel_limit=None,
    cluster_pixel_scale=0.1,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_ldm__lens_bulge_disk_mlr_nfw__source_inversion"

    pipeline_tag = pipeline_tagging.pipeline_tag_from_pipeline_settings(
        include_shear=pipeline_settings.include_shear,
        align_bulge_disk_centre=pipeline_settings.align_bulge_disk_centre,
        align_bulge_disk_phi=pipeline_settings.align_bulge_disk_phi,
        align_bulge_disk_axis_ratio=pipeline_settings.align_bulge_disk_axis_ratio,
        disk_as_sersic=pipeline_settings.disk_as_sersic,
        align_bulge_dark_centre=pipeline_settings.align_bulge_dark_centre,
        pixelization=pipeline_settings.pixelization,
        regularization=pipeline_settings.regularization,
    )

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's light and mass and one source galaxy, where we:

    # 1) Pass priors on the lens galaxy's light using the EllipticalSersic and EllipticalExponential of the previous
    #    pipeline. This includes using the bulge-disk alignment assumed in that pipeline.
    # 2) Pass priors on the lens galaxy's SphericalNFW mass profile's centre using the EllipticalIsothermal fit of the
    #    previous pipeline, if the NFW centre is a free parameter.
    # 3) Pass priors on the lens galaxy's shear using the ExternalShear fit of the previous pipeline.
    # 4) Pass priors on the source galaxy's inversion using the Pixelization and Regularization of the previous pipeline.

    class LensSourcePhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            ### Lens Light to Light + Mass, Sersic -> Sersic ###

            self.galaxies.lens.bulge_mass.centre = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.bulge.centre

            self.galaxies.lens.bulge_mass.axis_ratio = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.bulge.axis_ratio

            self.galaxies.lens.bulge_mass.phi = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.bulge.phi

            self.galaxies.lens.bulge_mass.intensity = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.bulge.intensity

            self.galaxies.lens.bulge_mass.effective_radius = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.bulge.effective_radius

            self.galaxies.lens.bulge_mass.sersic_index = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.bulge.sersic_index

            ### Lens Light to Light + Mass, Bulge -> Bulge, Disk -> Disk ###

            self.galaxies.lens.disk_mass.centre = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.disk.centre

            self.galaxies.lens.disk_mass.axis_ratio = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.disk.axis_ratio

            self.galaxies.lens.disk_mass.phi = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.disk.phi

            self.galaxies.lens.disk_mass.intensity = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.disk.intensity

            self.galaxies.lens.disk_mass.effective_radius = results.from_phase(
                "phase_4__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.disk.effective_radius

            ### Lens Mass, SIE ->  NFW ###

            if pipeline_settings.align_bulge_dark_centre:

                self.galaxies.lens.dark.centre = self.galaxies.lens.bulge_mass.centre

            elif not pipeline_settings.align_bulge_dark_centre:

                self.galaxies.lens.dark.centre = (
                    results.from_phase("phase_4__lens_bulge_disk_sie__source_inversion")
                    .variable_absolute(a=0.05)
                    .galaxies.lens.bulge_mass.centre
                )

            ### Lens Shear, Shear -> Shear ###

            self.galaxies.lens.shear = results.from_phase(
                "phase_2__lens_bulge_disk_sie__source_inversion"
            ).variable.galaxies.lens.shear

            ### Source Inversion, Inv -> Inv ###

            self.galaxies.source = results.from_phase(
                "phase_2__lens_bulge_disk_sie__source_inversion"
            ).inversion.constant.galaxies.source

    phase1 = LensSourcePhase(
        phase_name="phase_1__lens_bulge_disk_mlr_nfw__source_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                bulge_mass=lmp.EllipticalSersic,
                disk_mass=lmp.EllipticalExponential,
                mass=mp.EllipticalPowerLaw,
                shear=mp.ExternalShear,
            ),
            source=gm.GalaxyModel(
                redshift=redshift_source,
                pixelization=pipeline_settings.pixelization,
                regularization=pipeline_settings.regularization,
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        use_inversion_border=use_inversion_border,
        inversion_pixel_limit=inversion_pixel_limit,
        cluster_pixel_scale=cluster_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 75
    phase1.optimizer.sampling_efficiency = 0.2

    phase1 = phase1.extend_with_inversion_phase()

    return pipeline.PipelineImaging(pipeline_name, phase1)



import numpy as np
import pytest

from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data import convolution
from autolens.data.array import grids
from autolens.data.array import mask as msk
from autolens.lens import lens_data as ld
from autolens.model.inversion import convolution as inversion_convolution


@pytest.fixture(name="lens_data_7x7")
def make_lens_data_7x7(ccd_data_7x7, mask_7x7):
    return ld.LensData(ccd_data=ccd_data_7x7, mask=mask_7x7)


@pytest.fixture(name="lens_data_6x6")
def make_lens_data_6x6(ccd_data_6x6, mask_6x6):
    return ld.LensData(ccd_data=ccd_data_6x6, mask=mask_6x6)


class TestLensData(object):
    def test__attributes(self, ccd_data_7x7, lens_data_7x7):
        assert lens_data_7x7.pixel_scale == ccd_data_7x7.pixel_scale
        assert lens_data_7x7.pixel_scale == 1.0

        assert (lens_data_7x7.unmasked_image == ccd_data_7x7.image).all()
        assert (lens_data_7x7.unmasked_image == np.ones((7, 7))).all()

        assert (lens_data_7x7.unmasked_noise_map == ccd_data_7x7.noise_map).all()
        assert (lens_data_7x7.unmasked_noise_map == 2.0 * np.ones((7, 7))).all()

        assert (lens_data_7x7.psf == ccd_data_7x7.psf).all()
        assert (lens_data_7x7.psf == np.ones((3, 3))).all()

        assert lens_data_7x7.image_psf_shape == (3, 3)
        assert lens_data_7x7.inversion_psf_shape == (3, 3)

    def test__masking(self, lens_data_7x7):

        assert (lens_data_7x7.mask_1d == np.full(fill_value=False, shape=(9))).all()
        assert (lens_data_7x7.image_1d == np.ones(9)).all()
        assert (lens_data_7x7.noise_map_1d == 2.0 * np.ones(9)).all()

        assert (
            lens_data_7x7.mask_2d
            == np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                ]
            )
        ).all()

        assert (
            lens_data_7x7.image(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        assert (
            lens_data_7x7.noise_map(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

    def test__grid_stack(
        self, lens_data_7x7, grid_7x7, sub_grid_7x7, blurring_grid_7x7
    ):

        assert (lens_data_7x7.grid_stack.regular == grid_7x7).all()
        assert (lens_data_7x7.grid_stack.sub == sub_grid_7x7).all()
        assert (lens_data_7x7.grid_stack.blurring == blurring_grid_7x7).all()

    def test__interp_pixel_scale_input__grid_stack_include_interpolators(
        self, ccd_data_7x7, mask_7x7
    ):

        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, interp_pixel_scale=1.0
        )

        grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
            mask=mask_7x7, sub_grid_size=2, psf_shape=(3, 3)
        )
        new_grid_stack = grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(
            interp_pixel_scale=1.0
        )

        assert (lens_data_7x7.grid_stack.regular == new_grid_stack.regular).all()
        assert (
            lens_data_7x7.grid_stack.regular.interpolator.vtx
            == new_grid_stack.regular.interpolator.vtx
        ).all()
        assert (
            lens_data_7x7.grid_stack.regular.interpolator.wts
            == new_grid_stack.regular.interpolator.wts
        ).all()

        assert (lens_data_7x7.grid_stack.sub == new_grid_stack.sub).all()
        assert (
            lens_data_7x7.grid_stack.sub.interpolator.vtx
            == new_grid_stack.sub.interpolator.vtx
        ).all()
        assert (
            lens_data_7x7.grid_stack.sub.interpolator.wts
            == new_grid_stack.sub.interpolator.wts
        ).all()

        assert (lens_data_7x7.grid_stack.blurring == new_grid_stack.blurring).all()
        assert (
            lens_data_7x7.grid_stack.blurring.interpolator.vtx
            == new_grid_stack.blurring.interpolator.vtx
        ).all()
        assert (
            lens_data_7x7.grid_stack.blurring.interpolator.wts
            == new_grid_stack.blurring.interpolator.wts
        ).all()

    def test__cluster_pixel_scale_is_input__correct_cluster_bin_up_calculated__inversion_max_pixels_changes_bin_up(
        self, ccd_data_7x7, mask_7x7, grid_7x7
    ):
        ccd_data_7x7.pixel_scale = 1.0
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, cluster_pixel_scale=1.0
        )

        assert lens_data_7x7.cluster.bin_up_factor == 1
        assert (lens_data_7x7.mask_2d == lens_data_7x7.cluster.mask).all()
        assert (lens_data_7x7.cluster == grid_7x7).all()
        assert (
            lens_data_7x7.cluster.cluster_to_regular_all
            == np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8]])
        ).all()

        ccd_data_7x7.pixel_scale = 1.0
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, cluster_pixel_scale=1.9
        )

        assert lens_data_7x7.cluster.bin_up_factor == 1
        assert (lens_data_7x7.mask_2d == lens_data_7x7.cluster.mask).all()
        assert (
            lens_data_7x7.cluster.cluster_to_regular_all
            == np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8]])
        ).all()

        ccd_data_7x7.pixel_scale = 1.0
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, cluster_pixel_scale=2.0
        )
        assert lens_data_7x7.cluster.bin_up_factor == 2
        assert (
            lens_data_7x7.cluster.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()
        assert (
            lens_data_7x7.cluster
            == np.array([[1.0, -1.0], [1.0, 1.0], [-1.0, -1.0], [-1.0, 1.0]])
        ).all()
        assert (
            lens_data_7x7.cluster.cluster_to_regular_all
            == np.array([[0, -1, -1, -1], [1, 2, -1, -1], [3, 6, -1, -1], [4, 5, 7, 8]])
        ).all()

        ccd_data_7x7.pixel_scale = 2.0
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, cluster_pixel_scale=1.0
        )

        assert lens_data_7x7.cluster.bin_up_factor == 1

        ccd_data_7x7.pixel_scale = 1.0
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, cluster_pixel_scale=None
        )

        assert lens_data_7x7.cluster.bin_up_factor == 1
        assert (lens_data_7x7.mask_2d == lens_data_7x7.cluster.mask).all()
        assert (lens_data_7x7.cluster == grid_7x7).all()
        assert (
            lens_data_7x7.cluster.cluster_to_regular_all
            == np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8]])
        ).all()

        ccd_data_7x7.pixel_scale = 1.0
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7,
            mask=mask_7x7,
            cluster_pixel_scale=2.0,
            cluster_pixel_limit=5,
        )

        assert lens_data_7x7.cluster.bin_up_factor == 1
        assert (lens_data_7x7.mask_2d == lens_data_7x7.cluster.mask).all()
        assert (
            lens_data_7x7.cluster.cluster_to_regular_all
            == np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8]])
        ).all()

    def test__border(self, lens_data_7x7):
        assert (lens_data_7x7.border == np.array([0, 1, 2, 3, 5, 6, 7, 8])).all()

    def test__convolvers(self, lens_data_7x7):
        assert type(lens_data_7x7.convolver_image) == convolution.ConvolverImage
        assert (
            type(lens_data_7x7.convolver_mapping_matrix)
            == inversion_convolution.ConvolverMappingMatrix
        )

    def test__inversion_max_pixels(self, ccd_data_7x7, mask_7x7):
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, cluster_pixel_limit=2
        )

        assert lens_data_7x7.cluster_pixel_limit == 2

        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, cluster_pixel_limit=5
        )

        assert lens_data_7x7.cluster_pixel_limit == 5

    def test__hyper_noise_map_max(self, ccd_data_7x7, mask_7x7):
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, hyper_noise_map_max=10.0
        )

        assert lens_data_7x7.hyper_noise_map_max == 10.0

        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, hyper_noise_map_max=20.0
        )

        assert lens_data_7x7.hyper_noise_map_max == 20.0

    def test__uses_inversion__does_not_create_mapping_matrix_conovolver_if_false(
        self, ccd_data_7x7, mask_7x7
    ):
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7, mask=mask_7x7, uses_inversion=False
        )

        assert lens_data_7x7.convolver_mapping_matrix == None

    def test__uses_cluster_inversion__does_not_create_cluster_grid_if_false(
        self, ccd_data_7x7, mask_7x7
    ):
        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data_7x7,
            mask=mask_7x7,
            cluster_pixel_scale=1.0,
            cluster_pixel_limit=1,
            uses_cluster_inversion=False,
        )

        assert lens_data_7x7.cluster == None
        assert lens_data_7x7.cluster_pixel_scale == None
        assert lens_data_7x7.cluster_pixel_limit == None

    def test__different_ccd_data_without_mock_objects__customize_constructor_inputs(
        self
    ):

        psf = abstract_data.PSF(np.ones((7, 7)), 1)
        ccd_data = ccd.CCDData(
            np.ones((19, 19)),
            pixel_scale=3.0,
            psf=psf,
            noise_map=2.0 * np.ones((19, 19)),
        )
        mask = msk.Mask.unmasked_for_shape_and_pixel_scale(
            shape=(19, 19), pixel_scale=1.0, invert=True
        )
        mask[9, 9] = False

        lens_data_7x7 = ld.LensData(
            ccd_data=ccd_data,
            mask=mask,
            sub_grid_size=8,
            image_psf_shape=(7, 7),
            inversion_psf_shape=(3, 3),
            positions=[np.array([[1.0, 1.0]])],
        )

        assert (lens_data_7x7.unmasked_image == np.ones((19, 19))).all()
        assert (lens_data_7x7.unmasked_noise_map == 2.0 * np.ones((19, 19))).all()
        assert (lens_data_7x7.psf == np.ones((7, 7))).all()

        assert lens_data_7x7.sub_grid_size == 8
        assert lens_data_7x7.convolver_image.psf_shape == (7, 7)
        assert lens_data_7x7.convolver_mapping_matrix.psf_shape == (3, 3)
        assert (lens_data_7x7.positions[0] == np.array([[1.0, 1.0]])).all()

        assert lens_data_7x7.image_psf_shape == (7, 7)
        assert lens_data_7x7.inversion_psf_shape == (3, 3)

    def test__lens_data_7x7_with_modified_image(self, lens_data_7x7):

        lens_data_7x7 = lens_data_7x7.new_lens_data_with_modified_image(
            modified_image=8.0 * np.ones((7, 7))
        )

        assert (lens_data_7x7.unmasked_image == 8.0 * np.ones((7, 7))).all()

        assert (lens_data_7x7.image_1d == 8.0 * np.ones(9)).all()

        assert (
            lens_data_7x7.image(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 8.0, 8.0, 8.0, 0.0, 0.0],
                    [0.0, 0.0, 8.0, 8.0, 8.0, 0.0, 0.0],
                    [0.0, 0.0, 8.0, 8.0, 8.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

    def test__lens_data_6x6_with_binned_up_ccd_data(self, lens_data_6x6):
        binned_up_psf = lens_data_6x6.ccd_data.psf.new_psf_with_rescaled_odd_dimensioned_array(
            rescale_factor=0.5
        )

        lens_data_6x6 = lens_data_6x6.new_lens_data_with_binned_up_ccd_data_and_mask(
            bin_up_factor=2
        )

        assert (lens_data_6x6.unmasked_image == np.ones((3, 3))).all()
        assert (lens_data_6x6.psf == binned_up_psf).all()
        assert (lens_data_6x6.unmasked_noise_map == np.ones((3, 3))).all()
        assert (
            lens_data_6x6.ccd_data.background_noise_map == 1.5 * np.ones((3, 3))
        ).all()
        assert (lens_data_6x6.ccd_data.poisson_noise_map == 2.0 * np.ones((3, 3))).all()
        assert (
            lens_data_6x6.ccd_data.exposure_time_map == 20.0 * np.ones((3, 3))
        ).all()
        assert (
            lens_data_6x6.ccd_data.background_sky_map == 6.0 * np.ones((3, 3))
        ).all()

        assert (
            lens_data_6x6.mask_2d
            == np.array([[True, True, True], [True, False, True], [True, True, True]])
        ).all()

        assert (lens_data_6x6.image_1d == np.ones((1))).all()
        assert (lens_data_6x6.noise_map_1d == np.ones((1))).all()

    def test__lens_data_7x7_with_signal_to_noise_limit(
        self, ccd_data_7x7, lens_data_7x7
    ):

        lens_data_snr_limit = lens_data_7x7.new_lens_data_with_signal_to_noise_limit(
            signal_to_noise_limit=0.25
        )

        assert lens_data_snr_limit.pixel_scale == ccd_data_7x7.pixel_scale
        assert lens_data_snr_limit.pixel_scale == 1.0

        assert (lens_data_snr_limit.unmasked_image == ccd_data_7x7.image).all()
        assert (lens_data_snr_limit.unmasked_image == np.ones((7, 7))).all()

        assert (lens_data_snr_limit.unmasked_noise_map == 4.0 * np.ones((7, 7))).all()

        assert (lens_data_snr_limit.psf == ccd_data_7x7.psf).all()
        assert (lens_data_snr_limit.psf == np.ones((3, 3))).all()

        assert lens_data_snr_limit.image_psf_shape == (3, 3)
        assert lens_data_snr_limit.inversion_psf_shape == (3, 3)

        assert (lens_data_snr_limit.image_1d == np.ones(9)).all()
        assert (lens_data_snr_limit.noise_map_1d == 4.0 * np.ones(9)).all()

        assert (
            lens_data_snr_limit.noise_map(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 4.0, 4.0, 4.0, 0.0, 0.0],
                    [0.0, 0.0, 4.0, 4.0, 4.0, 0.0, 0.0],
                    [0.0, 0.0, 4.0, 4.0, 4.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()



import matplotlib.pyplot as plt

# plt.plot([1.2, 1.3, 1.4, 1.5], [0.97322, 0.96319, 0.95462, 0.94729])
# plt.plot([1.2, 1.3, 1.4, 1.5], [0.94228, 0.92098, 0.90311, 0.88790])

# slopes = [1.5, 1.4, 1.3]
# axis_ratios = [0.5, 0.5, 0.5]
# values = [0.90856, 0.92109, 0.93576]
#
# # values_calc = list(map(lambda slope, axis_ratio : (2.0*slope)**0.5 * axis_ratio, slopes, axis_ratios))
# values_calc = list(map(lambda slope, axis_ratio : 0.5*(slope / 0.5)**0.5**0.5, slopes, axis_ratios))
#
# print(values)
# print(values_calc)
# print(list(map(lambda value, value_calc : value - value_calc, values, values_calc)))
# # plt.show()

slopes = [1.5, 0.5]
axis_ratios = [0.5, 0.5]
values = [0.86602, 1.1547]


# values_calc = list(map(lambda slope, axis_ratio : (2.0*slope)**0.5 * axis_ratio, slopes, axis_ratios))
values_calc = list(
    map(
        lambda slope, axis_ratio: (1 - axis_ratio ** 2.0) ** (slope - 1),
        slopes,
        axis_ratios,
    )
)

print(values)
print(values_calc)
print(list(map(lambda value, value_calc: value - value_calc, values, values_calc)))
# plt.show()
